module CFG where

import Control.Arrow
import Control.Monad
import Data.Char
import qualified Data.Set as S
import Data.Set (Set)
import qualified Data.Map as M
import Data.Map (Map)
import Data.List
import Data.Maybe
import Data.Ord
import System.IO

data Symbol = Epsilon | Terminal String | Nonterminal String
  deriving (Eq, Ord)

data Production = Symbol :->: [Symbol]
  deriving (Eq, Ord)
  
data CFG = CFG {nonterminals :: Set String, alphabet :: Set String, startSymbol :: String, productions :: Set Production}
  deriving (Ord, Eq)

startWord :: CFG -> Symbol
startWord g = Nonterminal (startSymbol g)

trivialProduction :: CFG -> Production
trivialProduction g = startWord g :->: []

apply :: Production -> [Symbol] -> [[Symbol]]
apply (lhs :->: rhs) [] = []
apply p@(lhs :->: rhs) (x : xs) = map (x:) (apply p xs) ++
  case lhs == x of
    False  -> []
    True -> [rhs ++ xs]
    
mapMaybeSet :: (Ord a, Ord b) => (a -> Maybe b) -> Set a -> Set b
mapMaybeSet f = S.map fromJust . S.delete Nothing . S.map f
    
toWord :: [Symbol] -> Maybe [String]
toWord = (mapM (\x -> case x of {Terminal s -> Just s; _ -> Nothing}))
            


-----------------------------------------------------------------------------------------------------------------------
-- Pretty printing
-----------------------------------------------------------------------------------------------------------------------

instance Show Symbol where
  show (Terminal c) = c
  show (Nonterminal c) = c
  show Epsilon = ""

showWord :: [Symbol] -> String
showWord xs = "\"" ++ concatMap show xs ++ "\""

showWord' :: String -> String
showWord' xs = "\"" ++ xs ++ "\""

instance Show Production where
  show (lhs :->: rhs) = show lhs ++ " -> " ++ concat (intersperse " " (map show rhs))


-----------------------------------------------------------------------------------------------------------------------
-- Parsing
-----------------------------------------------------------------------------------------------------------------------

trim :: String -> String
trim = reverse . dropWhile isSpace . reverse . dropWhile isSpace

split :: Char -> String -> [String]
split c s = case span (/= c) s of
            (s', []) -> [s']
            (s', _ : s'') -> s' : split c s''

parseCommaSep :: String -> [String]
parseCommaSep = map trim . split ','

parseSymbol :: [String] -> [String] -> String -> Symbol
parseSymbol nts alph s
  | s == "" = Epsilon
  | s `elem` nts = Nonterminal s
  | s `elem` alph = Terminal s
  | otherwise = error ("Invalid symbol: '" ++ s ++ "'")

isValidProduction :: [String] -> [String] -> String -> Production -> Bool
isValidProduction nts alph s p@(lhs :->: rhs)
  | parseSymbol nts alph s /= Nonterminal s = False 
  | otherwise = 1 <= length rhs

parseProduction :: [String] -> [String] -> String -> [Production]
parseProduction nts alph s =
  case span (\c -> c /= '-' && c /= '→') s of
    (lhs, '-' : '>' : rhs) -> go lhs rhs
    (lhs, '→' : rhs) -> go lhs rhs
    _ -> err 
  where err = error $ "Failed to parse production: '" ++ s ++ "'"
        go lhs rhs = 
            case trim lhs of
                s -> if (trim rhs) == "" then [parseSymbol nts alph s :->: [Epsilon]]
                     else map (\rhs -> parseSymbol nts alph s :->: if null (trim rhs) then [Epsilon] else (map (parseSymbol nts alph) (split ' ' (trim rhs)))) (split '|' rhs)

cfg :: Set String -> Set String -> String -> Set Production -> CFG
cfg nts alph s ps 
  | not (s `S.member` nts) = 
      error $ "Illegal start symbol '" ++ s ++ "'" 
  | not (all valid (S.toList ps)) =
      error $ "Illegal production: '" ++ show (head (filter (not . valid) (S.toList ps)))
  | otherwise = CFG nts alph s ps
  where valid = isValidProduction (S.toList nts) (S.toList alph) s

parseCFG :: [String] -> CFG
parseCFG ("Grammar" : ntLine : alphLine : startLine : "Productions:" : ps) =
  case fmap trim (stripPrefix "Startsymbol:" startLine) of
    Just s -> 
        case fmap trim (stripPrefix "Nonterminals:" ntLine) of
            Just nts -> 
                case fmap trim (stripPrefix "Alphabet:" alphLine) of
                    Just alph -> let ntlist = parseCommaSep nts 
                                     alphlist = parseCommaSep alph
                                 in cfg (S.fromList $ ntlist) (S.fromList $ alphlist) (trim s) (S.fromList $ concatMap (parseProduction ntlist alphlist) ps)
                    _ -> error $ "Expected 'Alphabet:' but got '" ++ alphLine ++ "' instead."
            _ -> error $ "Expected 'Nonterminals:' but got '" ++ ntLine ++ "' instead."
    _ -> error $ "Expected 'Start symbol' but got '" ++ startLine ++ "' instead."
parseCFG _ = error "Could not parse CFG definition"

instance Show CFG where
  show (CFG nts alph s ps) = "Grammar\nNonterminals: " ++ intercalate "," (S.toList nts) ++ "\nAlphabet: " ++ intercalate "," (S.toList alph) ++ "\nStartsymbol: " ++ s ++ "\nProductions:\n" ++ intercalate "\n" (sort (map show (S.toList ps))) ++ "\nEND"

