module PDA (PDA(..), printPDA, parsePDA) where

import Data.List
import qualified Data.Set as S
import Data.Set (Set)
import Data.Char

data PDA s c = PDA {states :: Set s, alphabet :: Set c, stackAlphabet :: Set c,
                    transitions :: Set (s, c, c, s, [c]), startState :: s, startSymbol :: c}
  deriving Show

-- Printing

printPDA :: PDA String Char -> IO ()
printPDA m =
  do putStrLn "PDA"
     putStrLn $ "Alphabet: " ++ intersperse ';' (S.toList (alphabet m))
     putStrLn $ "States: " ++ concat (intersperse ";" (S.toList (states m)))
     putStrLn $ "Init: " ++ startState m
     putStrLn $ "Stackalphabet: " ++ intersperse ';' (S.toList (stackAlphabet m))
     putStrLn $ "Startsymbol: " ++ [startSymbol m] 
     putStrLn "Transitions:"
     mapM_ (putStrLn . showTransition) (S.toList (transitions m))
     putStrLn "END"
  where showTransition (q, c, z, q', zs) = q ++ ";" ++ [c] ++ ";" ++ [z] ++ ";" ++ q' ++ ";" ++ zs

parsePDA :: [String] -> PDA String Char
parsePDA (pda : alph : st : init : stalph : sts : tr : trs)
  | trim pda /= "PDA" = error $ "Expected token 'PDA', but got '" ++ trim pda ++ "'"
  | trim tr /= "Transitions:" = error $ "Expected token 'Transitions', but got '" ++ trim tr ++ "'"
  | otherwise = PDA {states = parseStates st, alphabet = parseAlphabet alph, startState = parseInit init,
                     stackAlphabet = parseStackAlphabet stalph, startSymbol = parseStartSymbol sts, transitions = S.fromList (map parseTransition trs)}
  where trim = reverse . dropWhile isSpace . reverse . dropWhile isSpace
        stripLabel s1 s2 =
          case stripPrefix (s1 ++ ":") s2 of
            Nothing -> error ("Expected entry '" ++ s1 ++ "', but got '" ++ s2 ++ "'")
            Just s  -> trim s

        splitOn :: Eq a => a -> [a] -> [[a]]
        splitOn _ [] = []
        splitOn c s  = case span (/= c) s of
                         (s1, []) -> [s1]
                         (s1, _ : s2) -> s1 : splitOn c s2

        parseChar [c] = c
        parseChar s   = error $ "Expected single character, but got '" ++ s ++ "'"
        parseLabel s = if null s then 'ε' else parseChar s
        parseAlphabet = S.fromList . map (parseChar . trim) . splitOn ';' . stripLabel "Alphabet"
        parseStackAlphabet = S.fromList . map (parseChar . trim) . splitOn ';' . stripLabel "Stackalphabet"
        parseStates = S.fromList . map trim . splitOn ';' . stripLabel "States"
        parseInit = trim . (stripLabel "Init")
        parseStartSymbol = parseChar . trim . stripLabel "Startsymbol"
        parseTransition s =
          case map trim (splitOn ';' s) of
            [a, b, c, d] -> (a, parseLabel b, parseChar c, d, [])
            [a, b, c, d,e] -> (a, parseLabel b, parseChar c, d, e)
            _         -> error $ "Invalid transition: '" ++ s ++ "'"                    
parsePDA _ = error "Invalid PDA format"

