import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.TreeSet;
import java.util.stream.Collectors;

/**
 * Provides the table data structure needed for the CYK algorithm, i.e. a table with entries for any indices
 * i,j with 1 ≤ i ≤ j ≤ n. 
 */
public class CYKTable {

	private final Set[][] table;
	private final int size;

	public CYKTable(int size) {
		if (size == 0) {
			throw new IllegalArgumentException("Zero size");
		}
		this.size = size;
		this.table = new Set[size][];
		for (int rowIndex = 0; rowIndex < size; rowIndex++) {
			Set[] row = new Set[size - rowIndex];
			Arrays.setAll(row, k -> new HashSet());
			table[rowIndex] = row;
		}
	}

	private void checkIndex(int i, int j) {
		if (i <= 0 || size < i) {
			throw new IllegalArgumentException("Invalid index i " + i);
		}
		if (j < i || size < j) {
			throw new IllegalArgumentException("Invalid index j " + j);
		}
	}

	/**
	 * Retrieves the set of derivation trees stored for the given indices.
	 */
	Set<DerivationTree> get(int i, int j) {
		checkIndex(i, j);
		Set<?> cell = table[j - i][i - 1];
		return (Set<DerivationTree>) cell;
	}

	/**
	 * Stores a set of derivation trees for the given indices.
	 */
	void set(int i, int j, Set<DerivationTree> trees) {
		checkIndex(i, j);
		table[j - i][i - 1] = new HashSet(Objects.requireNonNull(trees));
	}

	/**
	 * Adds a tree to the set of trees for the given indices.
	 */
	void add(int i, int j, DerivationTree tree) {
		checkIndex(i, j);
		table[j - i][i - 1].add(tree);
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder(size * size);
		for (int rowIndex = size - 1; rowIndex >= 0; rowIndex--) {
			Set[] row = table[rowIndex];
			assert row.length == size - rowIndex;
			for (Set cell : row) {
				String rowString = String.join(",", Util.collectionToString(new TreeSet<>(cell)));
				builder.append("{").append(rowString).append("}");
			}
			if (rowIndex > 0) {
				builder.append("\n");
			}
		}
		return builder.toString();
	}
}