import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Scanner;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

class ContextFreeGrammar {

	private static final Pattern TRANSITION_PATTERN = Pattern.compile(" -> *");
	private static final Pattern PRODUCTION_PATTERN = Pattern.compile(" +");

	public final Set<Terminal> alphabet;
	public final Set<NonTerminal> nonTerminals;
	public final Set<Production> productions;
	public final NonTerminal startSymbol;
	
	private final Map<Symbol, Set<List<Symbol>>> productionMap;
	private final Map<List<Symbol>, Set<NonTerminal>> reverseProductionMap;

	/**
	 * Construct a CFG with the given parameters.
	 */
	public ContextFreeGrammar(Set<Terminal> alphabet, Set<NonTerminal> nonTerminals, Set<Production> productions,
			NonTerminal startSymbol) {
		this.alphabet = Collections.unmodifiableSet(new HashSet<>(Objects.requireNonNull(alphabet)));
		this.nonTerminals = Collections.unmodifiableSet(new HashSet<>(Objects.requireNonNull(nonTerminals)));
		this.productions = Collections.unmodifiableSet(new HashSet<>(Objects.requireNonNull(productions)));
		this.startSymbol = startSymbol;
		checkValidGrammar();
		
		this.productionMap = new HashMap<>();
		this.reverseProductionMap = new HashMap<>();
		nonTerminals.stream().forEach(c -> productionMap.put(c, new HashSet<>()));
		for (Production p : productions) {
			productionMap.get(p.left).add(p.right);
			Set<NonTerminal> s = reverseProductionMap.get(p.right);
			if (s == null) {
				s = new HashSet<>();
				reverseProductionMap.put(p.right, s);
			}
			s.add(p.left);
		}
	}

	/**
	 * Parsing. You can ignore this.
	 */
	static ContextFreeGrammar parse(Scanner scanner) {
		// First line
		String first = scanner.nextLine().trim();
		if (!first.equals("Grammar")) {
			throw new IllegalArgumentException("Parsed grammar does not start with 'Grammar'.");
		}

		// Second line: Non-Terminals
		String nonTerminalsString = scanner.nextLine();
		if (!nonTerminalsString.startsWith("Nonterminals:")) {
			throw new IllegalArgumentException("Parsed grammar does not declare Nonterminals first.");
		}
		nonTerminalsString = nonTerminalsString.substring("Nonterminals:".length()).trim();

		Set<NonTerminal> nonTerminals = new HashSet<>();
		for (String s : nonTerminalsString.split(","))
			nonTerminals.add(new NonTerminal(s));

		// Third line: Alphabet
		String alphabetString = scanner.nextLine();
		if (!alphabetString.startsWith("Alphabet:")) {
			throw new IllegalArgumentException("Parsed grammar does not declare Alphabet second.");
		}
		alphabetString = alphabetString.substring("Alphabet:".length()).trim();

		Set<Character> alphabetLetters = new HashSet<>();
		for (String terminal : alphabetString.split(",")) {
			if (terminal.length() == 1) {
				alphabetLetters.add(terminal.charAt(0));
			} else {
				throw new IllegalArgumentException("Alphabet has to be input as a comma separated list "
						+ "without spaces. Terminals may only be chars.");
			}
		}

		// Fourth line; Startsymbol
		String initialName = scanner.nextLine();
		if (!initialName.startsWith("Startsymbol:")) {
			throw new IllegalArgumentException("Parsed grammar does not declare start symbol third.");
		}
		initialName = initialName.substring("Startsymbol:".length()).trim();
		NonTerminal startSymbol = new NonTerminal(initialName);

		// Productions
		if (!scanner.nextLine().equals("Productions:")) {
			throw new IllegalArgumentException("Parsed grammar does not declare productions");
		}

		Set<Production> productions = new HashSet<>();
		String production;
		while (!(production = scanner.nextLine()).equals("END")) {
			if (!production.contains(" ->")) {
				throw new IllegalArgumentException("Production " + production + " does not contain ' ->'");
			}
			String[] split = TRANSITION_PATTERN.split(production);
			if (split.length > 2) {
				throw new IllegalArgumentException("Invalid production " + production);
			}

			if (split[0].trim().contains(" "))
				throw new IllegalArgumentException("Left-hand side must contain exactly one symbol.");
			NonTerminal left = new NonTerminal(split[0].trim());

			if (split.length == 1) { // "A -> "; empty production.
				productions.add(new Production(left, Collections.emptyList()));
			} else {
				String[] right = split[1].split("\\|");
				for (String r : right) {
					r = r.trim();
					String[] rightSideProductions = PRODUCTION_PATTERN.split(r);
					List<Symbol> atoms = new ArrayList<>(rightSideProductions.length);
					for (String atom : rightSideProductions) {
						atom = atom.trim();
						if (atom.length() == 1 && alphabetLetters.contains(atom.charAt(0))) {
							atoms.add(new Terminal(atom.charAt(0)));
						} else if (!atom.isEmpty()) {
							atoms.add(new NonTerminal(atom));
						}
					}
					productions.add(new Production(left, atoms));
				}
				if (split[1].startsWith("|") || split[1].endsWith("|")) {
					productions.add(new Production(left));
				}
			}
		}

		Set<Terminal> alphabet = alphabetLetters.stream().map(Terminal::new).collect(Collectors.toSet());
		return new ContextFreeGrammar(alphabet, nonTerminals, productions, startSymbol);
	}

	/**
	 * Returns the set of all productions whose left-hand side consists of the given symbol.
	 */
	public Set<List<Symbol>> getProductionsByLhs(Symbol lhs) {
		Set<List<Symbol>> s = productionMap.get(lhs);
		if (s == null)
			return Collections.emptySet();
		else
			return s;
	}

	/**
	 * Returns the set of all productions with the given right-hand side.
	 */
	public Set<NonTerminal> getProductionsByRhs(List<Symbol> rhs) {
		Set<NonTerminal> s = reverseProductionMap.get(rhs);
		if (s == null)
			return Collections.emptySet();
		else
			return s;
	}

	/**
	 * Returns the set of all productions with the given right-hand side.
	 */
	public Set<NonTerminal> getProductionsByRhs(Symbol... rhs) {
		return getProductionsByRhs(List.of(rhs));
	}

	/**
	 * Printing. You can ignore this.
	 */
	public String toString() {
		String productionsString = productions.isEmpty() ? ""
				: String.join("\n", Util.collectionToString(new TreeSet<>(productions))) + "\n";
		return "Grammar\n" + "Nonterminals: " + String.join(",", Util.collectionToString(new TreeSet<>(nonTerminals)))
				+ "\n" + "Alphabet: " + String.join(",", Util.collectionToString(new TreeSet<>(alphabet))) + "\n"
				+ "Startsymbol: " + startSymbol + "\n" + "Productions:\n" + productionsString + "END";
	}
	
	public boolean isCNF() {
		return productions.stream().allMatch(p -> 
				 p.right.size() == 1 && alphabet.contains(p.right.get(0)) ||
				 p.right.size() == 2 && p.right.stream().allMatch(x -> nonTerminals.contains(x)));
	}

	/**
	 * Checking. You can ignore this.
	 */
	private void checkValidGrammar() {
		for (Terminal t : alphabet) {
			for (NonTerminal nonTerminal : nonTerminals) {
				if (String.valueOf(t.c).equals(nonTerminal.getName())) {
					throw new IllegalArgumentException(
							"Terminals and non-terminals must be disjoint! " + t + " is part of both.");
				}
			}
		}
		if (!nonTerminals.contains(startSymbol)) {
			throw new IllegalArgumentException("Start symbol " + startSymbol + " must be part of the non-terminals");
		}

		Set<Symbol> atoms = new HashSet<>(nonTerminals);
		atoms.addAll(alphabet);

		for (Production p : productions) {
			if (!atoms.contains(p.left))
				throw new IllegalArgumentException(p.left + " is not in the grammar");
			for (Symbol atom : p.right) {
				if (!atoms.contains(atom))
					throw new IllegalArgumentException(atom + " is not in the grammar");
			}
		}
	}

}